/* Copyright (C) 1995, 1996 Zlatko Calusic <maverick@fly.cc.fer.hr>
This file is part of DownLoad, FTP Client

The DownLoad is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation.

The DownLoad is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
the DownLoad source; see the file COPYING.  If not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include "dl.h"

extern int master, control, data, local;

extern int nofatal;

extern int aborted, wildcard;

extern char host[], file[], buffer[], *matching;

extern char localaddr[];

extern struct hostent *hptr;

extern struct sockaddr_in data_in;
extern int addrlen;

extern fd_set fdset;

extern struct timeval ttimer;

extern struct stat statbuf;

int get_response(void)
{
	int size;
	char *pos, *cptr = buffer, *where = buffer;

	for (;;)
	{
		timer_on(control);
		if (select(255, &fdset, NULL, NULL, &ttimer) <= 0)
		{
			fputs("Control connection timed out.\n", stderr);
			return ERROR;
		}
		if ((size = read(control, cptr, BUFFER)) < 0)
		{
			perror("Error reading control socket");
			return ERROR;
		}
		else if (!size)
		{
			fputs("Received EOF on control socket.\n", stderr);
			return ERROR;
		}
		cptr += size;
		*cptr = 0;
		while ((pos = strchr(where, '\n')))
		{
			strcpy(pos - 1, "\n");
			fputs(where + 4, stderr);
			if (*(where + 3) == ' ' && atoi(buffer) == atoi(where))
				switch (atoi(where) / 10)
				{
					case 50:
						if (nofatal)
							return ERROR;
					case 55:
						if (!stat(file, &statbuf) && !statbuf.st_size)
							unlink(file);
						fputs("Aborting download ...\n", stderr);
						aborted = 1;
						return ERROR;
					case 12:
					case 42:
					case 45:
					case 53:
						return ERROR;
					default:
						return OK;
				}
			else
				where = pos + 1;
		}
	}
}

int ftprequest(char *format, char *arg)
{
	int length;
	char string[256];

	if (arg)
		sprintf(string, format, arg);
	else
		strcpy(string, format);
	length = strlen(string);
	if (write(control, string, length) != length)
		return ERROR;
	else
		return get_response();
}

int establish_connection()
{
	struct sockaddr_in comm_in;
	unsigned long laddr;

	if (control)
	{
		close(control);
		control = 0;
	}
	if ((control = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
		control = 0;
		perror("Error creating socket");
		return ERROR;
	}
	fprintf(stderr, "Connecting to %s ...", host);
	if ((int) (laddr = inet_addr(host)) == -1 || !strcmp(host, "0"))
		hptr = gethostbyname(host);
	else
		hptr = gethostbyaddr((char *) &laddr, sizeof(laddr), AF_INET);
	if (hptr)
		memcpy(&comm_in.sin_addr, hptr->h_addr_list[0], 4);
	else
	{
		fputs("\nCan't resolve remote hostname, aborting ...\n", stderr);
		aborted = 1;
		return ERROR;
	}
	comm_in.sin_family = AF_INET;
	comm_in.sin_port = htons(21);
	if (connect(control, (struct sockaddr *) &comm_in, sizeof(comm_in)) < 0)
	{
		perror("\nError connecting to server");
		return ERROR;
	}
	fputs(" connected.\n", stderr);
	return get_response();
}

int get_port(void)
{
	int port;
	char string[64];

	if (master)
	{
		close(master);
		master = 0;
	}
	if ((master = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
		master = 0;
		perror("Error creating socket");
		return ERROR;
	}
	data_in.sin_family = AF_INET;
	data_in.sin_addr.s_addr = htonl(0);
	data_in.sin_port = htons(0);
	addrlen = sizeof(data_in);
	if (bind(master, (struct sockaddr *) &data_in, addrlen) < 0)
	{
		perror("Error binding socket");
		return ERROR;
	}
	if (getsockname(master, (struct sockaddr *) &data_in, &addrlen) < 0)
	{
		perror("Error getting port");
		return ERROR;
	}
	port = ntohs(data_in.sin_port);
	if (listen(master, 1) < 0)
	{
		perror("Error listening on data port");
		return ERROR;
	}
	sprintf(string, "PORT %u,%u,%u,%u,%d,%d\r\n", (unsigned char) localaddr[0], (unsigned char) localaddr[1], (unsigned char) localaddr[2], (unsigned char) localaddr[3], (port & 0xff00) >> 8, port & 0xff);
	return ftprequest(string, NULL);
}

int accept_data(void)
{
	if (data)
	{
		close(data);
		data = 0;
	}
	addrlen = sizeof(data_in);
	if ((data = accept(master, (struct sockaddr *) &data_in, &addrlen)) < 0)
	{
		data = 0;
		perror("Error accepting connection on data port");
		return ERROR;
	}
	return OK;
}

int get_list(void)
{
	int size, header = 0;
	char *found, *fileptr, *cptr = buffer, *where = buffer;
	static char matched[BUFFER];

	matching = matched;
	*matching = 0;
	if (!get_port() || !ftprequest("LIST\r\n", NULL) || !accept_data())
		return ERROR;
	for (;;)
	{
		timer_on(data);
		if (select(255, &fdset, NULL, NULL, &ttimer) <= 0)
		{
			fputs("Data connection timed out.\n", stderr);
			return ERROR;
		}
		if ((size = read(data, cptr, BUFFER)) < 0)
		{
			perror("Error reading data socket");
			return ERROR;
		}
		else if (!size)
		{
			if (strlen(matching))
			{
				fputs("--------------------\n", stderr);
				wildcard = 1;
				strcpy(file, strtok(matching, "\r"));
				return get_response();
			}
			else
			{
				get_response();
				fputs("No matches found, aborting ...\n", stderr);
				aborted = 1;
				return ERROR;
			}
		}
		cptr += size;
		*cptr = 0;
		while ((found = strchr(where, '\n')))
		{
			*(found - 1) = 0;
			fileptr = strrchr(where, ' ') + 1;
			if (*where == '-' && match(file, fileptr, 1))
			{
				if (!header)
				{
					fputs("File(s) to download:\n--------------------\n", stderr);
					header = 1;
				}
				fprintf(stderr, "%s\n", fileptr);
				strcat(matching, fileptr);
				strcat(matching, "\r");
			}
			where = found + 1;
		}
	}
}

void timer_on(int fd)
{
	FD_ZERO(&fdset);
	FD_SET(fd, &fdset);
	ttimer.tv_sec = TMOUT;
	ttimer.tv_usec = 0;
	return;
}
